# IMPORTS ########################################################### IMPORTS #
# General Python imports
from __future__ import division
import scipy as sp
import scipy.linalg as spl
import cPickle
import multiprocessing as mp
from time import time
from matplotlib.mlab import find

# PyComm imports
import pycomm.modulation.psk as psk
import pycomm.auxiliary.berc as berc
import pycomm.reconstruction.subpur as subpur

# CLASSRECV_BER_TEST ###################################### CLASSRECV_BER_TEST #
def classrecv_ber_test(settings, DicMtx):
    N = settings.get('N')
    MeasMtx = sp.eye(N)
    return ber_test(MeasMtx, DicMtx, settings, whitening=0)

# CSRECV_BER_TEST ########################################### CSRECV_BER_TEST #
def csrecv_ber_test(settings, DicMtx):
    N = settings.get('N')
    kappa = settings.get('kappa')
    O = sp.int_(sp.round_(N/kappa))
    if O < 1:
        O = 1
    MeasMtx = 2*sp.random.randint(0,2,size=(O,N))-1
    return ber_test(MeasMtx, DicMtx, settings, whitening=1)

# RSRECV_DTPT_TEST ########################################### RSRECV_DTPT_TEST #
def rsrecv_ber_test(settings, DicMtx):
    N = settings.get('N')
    kappa = settings.get('kappa')
    O = sp.int_(sp.round_(N*kappa))
    if O < 1:
        O = 1
    I = sp.eye(N)
    row_idxs = sp.sort(sp.random.permutation(range(0,N))[0:O])
    MeasMtx = I[row_idxs,:]
    return ber_test(MeasMtx, DicMtx, settings, whitening=0)

# RDCSRECV_DTPT_TEST ##################################### RDCSRECV_DTPT_TEST #
def rdcsrecv_ber_test(settings, DicMtx):
    N = settings.get('N')
    kappa = settings.get('kappa')
    O = sp.int_(sp.round_(N/kappa))
    if O < 1:
        O = 1
    MeasVec = 2*sp.random.randint(0,2,N)-1
    MeasMtx = sp.zeros((O,N))
    ONratio = sp.int_(sp.floor(N/O))
    if ONratio < 1:
        ONratio = 1
    idx1 = 0; idx2 = 0
    for ii in xrange(O):
        idx1 = idx2
        if ii < sp.mod(N,O):
            idx2 = idx1+ONratio+1
            colIdx = sp.int_(sp.linspace(idx1, idx2, ONratio+1, endpoint=False))
        else:
            idx2 = idx1+ONratio
            colIdx = sp.int_(sp.linspace(idx1, idx2, ONratio, endpoint=False))
        MeasMtx[ii,colIdx] = MeasVec[colIdx]
    return ber_test(MeasMtx, DicMtx, settings, whitening=0)

# CSSCSRECV_DTPT_TEST ##################################### CSSCSRECV_DTPT_TEST #
def csscsrecv_ber_test(settings, DicMtx):
    N = settings.get('N')
    kappa = settings.get('kappa')
    O = sp.int_(sp.round_(N/kappa))
    if O < 1:
        O = 1
    MeasVec = sp.ones(N)
    MeasMtx = sp.zeros((O,N))
    ONratio = sp.int_(sp.floor(N/O))
    if ONratio < 1:
        ONratio = 1
    idx1 = 0; idx2 = 0
    for ii in xrange(O):
        idx1 = idx2
        if ii < sp.mod(N,O):
            idx2 = idx1+ONratio+1
            colIdx = sp.int_(sp.linspace(idx1, idx2, ONratio+1, endpoint=False))
        else:
            idx2 = idx1+ONratio
            colIdx = sp.int_(sp.linspace(idx1, idx2, ONratio, endpoint=False))
        MeasMtx[ii,colIdx] = MeasVec[colIdx]
    return ber_test(MeasMtx, DicMtx, settings, whitening=0)

# BER_TEST ######################################################### BER_TEST #
def ber_test(MeasMtx, DicMtx, settings, whitening=1):
    # Experiment settings
    SNR_dB = settings.get('SNR')
    modulationOrder = settings.get('modulationOrder')
    noRuns = settings.get('noRuns')
    noSymbols = settings.get('noSymbolsPerRun')
    N = settings.get('N')   # Number of Nyquist samples
    S = settings.get('S')   # Density, i.e. how many codes are in one signal
    noBits = sp.log2(modulationOrder)*S*noSymbols

    # Seed random generator
    if sp.size(mp.current_process()._identity)>0:
        procID = mp.current_process()._identity[0]
        sp.random.seed( int(1e8 * ( time()-int(time()) ) * procID) )
    
    errors = 0
    for ii in xrange(noRuns):
        # Generate bit vector
        bits = sp.random.randint(0,2,noBits)

        # Generate baseband CDMA signal
        x = generateSignal(bits, modulationOrder, N, S, DicMtx)
        
        # Add AWGN
        SNR_pwr = 10**(SNR_dB/10)
        P_s = sp.mean(abs(x)**2)
        P_n = P_s/SNR_pwr
        noise = sp.sqrt(P_n/2)*(sp.random.randn(x.size,1) + \
                                1j*sp.random.randn(x.size,1))
        y = x + noise

        # Receiver
        bits_hat = sp.zeros((noSymbols, sp.log2(modulationOrder)*S))
        for jj in xrange(noSymbols):
            # Isolate current symbol
            y_jj = y[jj*N:(jj+1)*N]

            # Sampling - the samples contain colored noise
            y_samp = sp.dot(MeasMtx, y_jj)

            # Pre-whitening
            if whitening:
                C = spl.cholesky(sp.dot(MeasMtx, MeasMtx.T), lower=True)
                y_samp = spl.solve(C, y_samp)
                MeasMtx = spl.solve(C, MeasMtx)

            # Demodulate
            symbBits = demodulateSignal(y_samp, MeasMtx, DicMtx, N,
                                        S, modulationOrder)
            bits_hat[jj,:] = symbBits

        bits_hat = sp.int_(bits_hat.reshape(bits.shape))

        # Calculate BER
        (ber, errors_ii) = berc(bits, bits_hat)
        errors += errors_ii
    return errors

# GENERATESIGNAL ############################################# GENERATESIGNAL #
def generateSignal(bits, modulationOrder, M, S, Dictionary):
    # Generate symbol vector
    (tmp, Constellation) = psk.mod(bits, modulationOrder, 0, S)

    # Randomly permute Alpha
    alpha = sp.zeros((M,tmp.shape[1]), complex)
    for ii in xrange(tmp.shape[1]):
        randomIdx = sp.sort(sp.random.permutation(sp.arange(0,M))[0:S])
        alpha[randomIdx,ii] = tmp[:,ii]

    # Apply CDMA modulation
    s_d = sp.dot(Dictionary, alpha)
    s_d = sp.reshape(s_d, (s_d.size, 1), order='F')
    
    return s_d

# DEMODULATESIGNAL ######################################### DEMODULATESIGNAL #
def demodulateSignal(s_d, MeasMtx, DicMtx, N, S, modulationOrder):
    # Generate A matrix
    A = sp.dot(MeasMtx, DicMtx)

    # Demodulate CDMA
    (x_hat, dmy1, dmy2) = subpur(s_d, A, S, False)
    
    # Identify significant codes and extract them
    threshold = sp.sort(abs(x_hat),0)[-S]
    alpha_hat = x_hat[find(abs(x_hat)>=threshold)][0:S]

    # Demodulate PSK
    bits_hat = psk.demod(alpha_hat, modulationOrder)
    return bits_hat
