# IMPORTS ########################################################### IMPORTS #
# General Python imports
from __future__ import division
import scipy as sp
import multiprocessing as mp
from time import time
from matplotlib.mlab import find

# PyComm imports
import pycomm.modulation.psk as psk
import pycomm.reconstruction.subpur as subpur

# CSRECV_DTPT_TEST ########################################### CSRECV_DTPT_TEST #
def csrecv_dtpt_test(settings, DicMtx):
    N = settings.get('N')
    kappa = settings.get('kappa')
    O = sp.int_(sp.round_(N*kappa))
    if O < 1:
        O = 1
    MeasMtx = 2*sp.random.randint(0,2,size=(O,N))-1
    return dtpt_test(MeasMtx, DicMtx, settings)

# RSRECV_DTPT_TEST ########################################### RSRECV_DTPT_TEST #
def rsrecv_dtpt_test(settings, DicMtx):
    N = settings.get('N')
    kappa = settings.get('kappa')
    O = sp.int_(sp.round_(N*kappa))
    if O < 1:
        O = 1
    I = sp.eye(N)
    row_idxs = sp.sort(sp.random.permutation(range(0,N))[0:O])
    MeasMtx = I[row_idxs,:]
    return dtpt_test(MeasMtx, DicMtx, settings)

# RDCSRECV_DTPT_TEST ########################################### RDCSRECV_DTPT_TEST #
def rdcsrecv_dtpt_test(settings, DicMtx):
    N = settings.get('N')
    kappa = settings.get('kappa')
    O = sp.int_(sp.round_(N*kappa))
    if O < 1:
        O = 1
    MeasVec = 2*sp.random.randint(0,2,N)-1
    MeasMtx = sp.zeros((O,N))
    ONratio = sp.int_(sp.floor(N/O))
    if ONratio < 1:
        ONratio = 1
    idx1 = 0; idx2 = 0
    for ii in xrange(O):
        idx1 = idx2
        if ii < sp.mod(N,O):
            idx2 = idx1+ONratio+1
            colIdx = sp.int_(sp.linspace(idx1, idx2, ONratio+1, endpoint=False))
        else:
            idx2 = idx1+ONratio
            colIdx = sp.int_(sp.linspace(idx1, idx2, ONratio, endpoint=False))
        MeasMtx[ii,colIdx] = MeasVec[colIdx]
    return dtpt_test(MeasMtx, DicMtx, settings)

# CSSCSRECV_DTPT_TEST ########################################### CSSCSRECV_DTPT_TEST #
def csscsrecv_dtpt_test(settings, DicMtx):
    N = settings.get('N')
    kappa = settings.get('kappa')
    O = sp.int_(sp.round_(N*kappa))
    if O < 1:
        O = 1
    MeasVec = sp.ones(N)
    MeasMtx = sp.zeros((O,N))
    ONratio = sp.int_(sp.floor(N/O))
    if ONratio < 1:
        ONratio = 1
    idx1 = 0; idx2 = 0
    for ii in xrange(O):
        idx1 = idx2
        if ii < sp.mod(N,O):
            idx2 = idx1+ONratio+1
            colIdx = sp.int_(sp.linspace(idx1, idx2, ONratio+1, endpoint=False))
        else:
            idx2 = idx1+ONratio
            colIdx = sp.int_(sp.linspace(idx1, idx2, ONratio, endpoint=False))
        MeasMtx[ii,colIdx] = MeasVec[colIdx]
    return dtpt_test(MeasMtx, DicMtx, settings)

# DTPT_TEST ######################################################### DTPT_TEST #
def dtpt_test(MeasMtx, DicMtx, settings):
    # Experiment settings
    modulationOrder = settings.get('modulationOrder')
    threshold = settings.get('threshold')
    N = settings.get('N')   # Number of Nyquist samples
    M = N                   # Number of CDMA codes used
    S = settings.get('S')   # Density, i.e. how many codes are in one signal
    noBits = sp.log2(modulationOrder)*S

    # Seed random generator
    if sp.size(mp.current_process()._identity)>0:
        procID = mp.current_process()._identity[0]
        sp.random.seed( int(1e8 * ( time()-int(time()) ) * procID) )
    
    # Generate bit vector
    bits = sp.random.randint(0,2,noBits)

    # Generate baseband CDMA signal
    x = sp.real(generateSignal(bits, modulationOrder, M, S, DicMtx))

    # Sampling
    y_samp = sp.dot(MeasMtx, x)

    # Demodulate
    (x_hat, k, proc_time) = demodulateSignal(y_samp, MeasMtx, DicMtx, N,\
                                S, modulationOrder)

    # Find MSE
    MSE = sp.mean((x-x_hat)**2)
    if MSE < threshold:
        success = 1
    else:
        success = 0
    return success, k, proc_time

# GENERATESIGNAL ############################################# GENERATESIGNAL #
def generateSignal(bits, modulationOrder, M, S, Dictionary):
    # Generate symbol vector
    (tmp, Constellation) = psk.mod(bits, modulationOrder, 0, S)

    # Randomly permute Alpha
    alpha = sp.zeros((M,tmp.shape[1]), complex)
    for ii in xrange(tmp.shape[1]):
        randomIdx = sp.sort(sp.random.permutation(sp.arange(0,M))[0:S])
        alpha[randomIdx,ii] = tmp[:,ii]

    # Apply CDMA modulation
    s_d = sp.dot(Dictionary, alpha)
    s_d = sp.reshape(s_d, (s_d.size, 1), order='F')

    return s_d

# DEMODULATESIGNAL ######################################### DEMODULATESIGNAL #
def demodulateSignal(s_d, MeasMtx, DicMtx, N, S, modulationOrder):
    # Generate A matrix
    A = sp.dot(MeasMtx, DicMtx)

    # Demodulate CDMA
    noRuns = 3
    for ii in xrange(noRuns): # Warmup
        (alpha_hat, dmy, k) = subpur(s_d, A, S, False)
    
    startTime = time()
    for ii in xrange(noRuns): 
        (alpha_hat, dmy, k) = subpur(s_d, A, S, False)
    proc_time = (time() - startTime)/noRuns

    # Reconstruct signal
    s_d_hat = sp.dot(DicMtx, alpha_hat)
    return s_d_hat, k, proc_time
