# IMPORTS ########################################################### IMPORTS #
# General Python imports
from __future__ import division
import scipy as sp
import cPickle
import os, sys
from time import time
from mpi4py import MPI

# PyComm imports
import pycomm.dictionaries.gold as gold

# My imports
import exp_funcs as ef
 
# Global variables
comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()
TERMINATE = sp.uint64(999999999)

# SERVER ################################################################# SERVER #
def server(recvStruc, m=5):
    # Experiment settings
    N = 2**m-1
    convergence_threshold = 10**-5
    threshold = 10**-6
    gridSize = 50
    settings = {'N': N,
                'S': 1,
                'kappa': 1,
                'modulationOrder': 2,
                'threshold': threshold,
                'gridSize': gridSize}
    resultsFileName = "dtpt_"+recvStruc+"_RESULTS.pkl"
    tmpFileName = "TMP_dtpt_"+recvStruc+"_RESULTS.pkl"
    
    # Stop-and-Resume initialization
    try:
        filePtr=open(tmpFileName,'rb')
        exp = cPickle.load(filePtr)
        filePtr.close()
        filePtr=open(resultsFileName,'rb')
        (settings, convergence, success, subpur_iters, proc_times) = cPickle.load(filePtr)
        success = success*exp
        subpur_iters = subpur_iters*exp
        proc_times = proc_times*exp
        filePtr.close()
    except IOError:
        exp = 0
        success = sp.zeros((gridSize, gridSize), dtype=sp.uint64)
        subpur_iters = sp.zeros((gridSize, gridSize))
        proc_times = sp.zeros((gridSize, gridSize))
        convergence = list()

    # Send settings and initial computations
    for ii in xrange(1,size):
        comm.send(settings, dest=ii, tag=77)
        comm.isend(gridSize, dest=ii, tag=77)

    # Main loop
    recvBuffer1 = sp.empty((gridSize, gridSize), dtype=sp.uint64)
    recvBuffer2 = sp.empty((gridSize, gridSize))
    recvBuffer3 = sp.empty((gridSize, gridSize))
    while True:
        # Receive results
        startTime = time()
        nodeId = comm.recv(source=MPI.ANY_SOURCE, tag=77)
        comm.Recv(recvBuffer1, source=nodeId, tag=77)
        comm.Recv(recvBuffer2, source=nodeId, tag=77)
        comm.Recv(recvBuffer3, source=nodeId, tag=77)
        old_success = success.copy()
        success += recvBuffer1
        subpur_iters += recvBuffer2
        proc_times += recvBuffer3
        exp += 1
        endTime = time()
        
        # Calculate convergence
        if exp == 1:
            MSE = sp.mean(success**2)
        else:
            MSE = sp.mean((old_success/(exp-1)-success/exp)**2)
        convergence.append(MSE)
        
        # Save results
        results = (settings, convergence, success/exp, subpur_iters/exp, proc_times/exp)
        filePtr = open(resultsFileName, 'wb')
        cPickle.dump(results, filePtr)
        filePtr.close()
        filePtr = open(tmpFileName, 'wb')
        cPickle.dump(exp, filePtr)
        filePtr.close()
        
        # Output printing
        print("#######################################")
        print("Received results from worker {}".format(nodeId))
        print("Experiment no.: {}".format(exp))
        print("Computation time: {} seconds".format(endTime-startTime))
        print("MSE: {}".format(MSE))
        print("")

        # Send new task or stop
        if MSE > convergence_threshold:
            comm.isend(gridSize, dest=nodeId, tag=77)
        else:
            break
    
    os.remove(tmpFileName)
    for ii in xrange(1,size):
        comm.isend(TERMINATE, dest=ii, tag=77)
    return

# CLIENT ############################################################# CLIENT #
def client(recvStruc):
    settings = comm.recv(source=0, tag=77)
    
    # Generate dictionary
    m = sp.int_(sp.log2(settings['N']+1))
    polys, initialConditions = getPolys(m)
    poly1 = sp.array(polys[0])
    poly2 = sp.array(polys[1])
    DicMtx = gold(m, poly1, poly2, initialConditions, True)

    while True:
        task = comm.recv(source=0, tag=77)
        if task == TERMINATE:
            break
        else:
            gridSize = task
            (success, subpur_iters, proc_times) = worker(recvStruc, gridSize, settings, DicMtx)
            comm.isend(rank, dest=0, tag=77)
            comm.Isend(success, dest=0, tag=77)
            comm.Isend(subpur_iters, dest=0, tag=77)
            comm.Isend(proc_times, dest=0, tag=77)

# WORKER ############################################################# WORKER #
def worker(recvStruc, gridSize, settings, DicMtx):
    success = sp.zeros((gridSize, gridSize), dtype=sp.uint64)
    subpur_iters = sp.zeros((gridSize, gridSize))
    proc_times = sp.zeros((gridSize, gridSize))
    for ii in xrange(1,sp.int_(gridSize)+1):
        kappa = ii/gridSize
        startTime = time()
        for jj in xrange(1,sp.int_(gridSize)+1):
            rho = jj/gridSize
            S = sp.int_(sp.round_(rho*kappa*settings['N']))
            if S < 1:
                S=1
            settings['S'] = S
            settings['kappa'] = kappa
            cmdString = "(success[jj-1,ii-1], subpur_iters[jj-1,ii-1], proc_times[jj-1,ii-1]) = ef."+recvStruc+"recv_dtpt_test(settings, DicMtx)"
            exec(cmdString)
        print("worker={}, kappa={}, time spent={}, N={}".format(rank, kappa, time()-startTime, settings['N']))
    return success, subpur_iters, proc_times

# GETPOLYS ######################################################### GETPOLYS #
def getPolys(m):
    polys = {
        3: list([[3, 1, 0], [3, 2, 0]]),
        5: list([[5, 2, 0], [5, 4, 3, 2, 0]]),
        7: list([[7, 6, 0], [7, 4, 0]]),
        10: list([[10, 3, 0], [10, 9, 8, 6, 3, 2, 0]])
        }
    
    initialConditions = sp.zeros((2,m))
    initialConditions[0,0]=1
    initialConditions[1,0]=1
    return polys[m], initialConditions

# MAIN ################################################################# MAIN #
def main(recvStruc, m):
    if rank == 0:
        # If we are on node 0 we will act as server
        server(recvStruc, m)
    else: 
        # "Workers" all nodes except the node with rank 0
        client(recvStruc)
