# IMPORTS ########################################################### IMPORTS #
# General Python imports
from __future__ import division
import scipy as sp
from matplotlib.mlab import find
import cPickle
import os, sys
from time import time
from mpi4py import MPI

# PyComm imports
import pycomm.dictionaries.gold as gold

# My imports
import exp_funcs as ef
 
# Global variables
comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()
TERMINATE = sp.uint64(999999999)

recvs = sp.array(['classrecv', 'csrecv', 'rdcsrecv', 'csscsrecv'])
activeRecvIdxs = sp.array([0, 3])
activeRecvs = recvs[activeRecvIdxs]
noActiveRecvs = activeRecvs.size

# SERVER ################################################################# SERVER #
def server(m, S, quanBits, kappa, noErrors, EbN0_Start, EbN0_End, expMask, noSymbolsPerRun):
    # Experiment settings
    N = 2**m-1
    noRunsPerTest = 10
    modulationOrder = 4
    EbN0s_dB = sp.linspace(EbN0_Start,EbN0_End,\
                int((EbN0_End-EbN0_Start)/3)+1, endpoint=True)
    noBitsPerTest = noRunsPerTest*noSymbolsPerRun*\
                        sp.log2(modulationOrder)*S
    settings = {'noRuns': noRunsPerTest,
                'N': N,
                'S': S,
                'modulationOrder': modulationOrder,
                'noSymbolsPerRun': noSymbolsPerRun,
                'kappa': kappa,
                'EbN0': 0,
                'quanBits': quanBits}
    resultsFileName = 'quan_rf_ber_main_RESULTS.pkl'
    tmpFileName = 'TMP_quan_rf_ber_main_RESULTS.pkl'
    global activeRecvs
    # Experiment mask - used to make simulation skip some 
    # experiment settings.
    # Rows -> EbN0-values and columns - > recv. strucs
    if expMask.shape != (len(EbN0s_dB),noActiveRecvs):
        raise(ValueError("expMask has wrong dimensions!"))
    
    # Send settings
    for ii in xrange(1,size):
        comm.send(settings, dest=ii, tag=77)
    
    # Simulation BER
    noEbN0s = len(EbN0s_dB)
    berSim = sp.zeros((noEbN0s, noActiveRecvs))
    
    # Stop-and-Resume initialization
    try:
        filePtr=open(tmpFileName,'rb')
        (done, errors, noBits) = cPickle.load(filePtr)
        filePtr.close()
        filePtr=open(resultsFileName,'rb')
        (settings, activeRecvs, EbN0s_dB, berSim, noBits) = cPickle.load(filePtr)
        filePtr.close()
    except IOError:
        done = sp.logical_not(expMask)
        errors = sp.zeros((noEbN0s, noActiveRecvs))
        noBits = sp.zeros((noEbN0s, noActiveRecvs))

    # Main loop
    for ii in xrange(noEbN0s):
        print("################################")
        while True:
            startTime = time()
            for jj in xrange(noActiveRecvs):
                if done[ii,jj]:
                    continue
                
                # Send tasks
                sendBuffer = sp.empty(2, dtype=sp.int64)
                for kk in xrange(1,size):
                    sendBuffer[0] = jj
                    sendBuffer[1] = EbN0s_dB[ii]
                    comm.Isend(sendBuffer, dest=kk, tag=77)

                # Receive tasks
                recvBuffer = sp.empty(1, dtype=sp.uint64)
                for kk in xrange(1,size):
                    comm.Recv(recvBuffer, source=kk, tag=77)
                    errors[ii,jj] += recvBuffer[0]
                    noBits[ii,jj] += noBitsPerTest
                
                berSim[ii,jj] = errors[ii,jj]/noBits[ii,jj]
                if errors[ii,jj] > noErrors:
                    done[ii,jj] = 1
            endTime = time()
            
            # Save results
            results = (settings, activeRecvs, EbN0s_dB, berSim, noBits)
            filePtr = open(resultsFileName, 'wb')
            cPickle.dump(results, filePtr)
            filePtr.close()
            backup = (done, errors, noBits)
            filePtr = open(tmpFileName, 'wb')
            cPickle.dump(backup, filePtr)
            filePtr.close()
            
            # Output printing
            print("EbN0: {:d} dB".format(int(EbN0s_dB[ii])))
            for jj in xrange(noActiveRecvs):
                print("  Recv. {:d}: Bits sent: {:d}, errors: {:d}".format(jj,\
                        int(noBits[ii,jj]), int(errors[ii,jj])))
            print("Computation time: {} seconds".format(endTime-startTime))
            print("")

            # Stop?
            if sp.sum(done[ii,:]) >= noActiveRecvs:
                break
    
    os.remove(tmpFileName)
    for kk in xrange(1,size):
        comm.Isend(TERMINATE, dest=kk, tag=77)
    return

# CLIENT ############################################################# CLIENT #
def client():
    sendBuffer = sp.empty(1, dtype=sp.uint64)
    task = sp.empty(2, dtype=sp.int64)
    
    settings = comm.recv(source=0, tag=77)
    
    # Generate dictionary
    m = sp.int_(sp.log2(settings['N']+1))
    polys, initialConditions = getPolys(m)
    poly1 = sp.array(polys[0])
    poly2 = sp.array(polys[1])
    DicMtx = gold(m, poly1, poly2, initialConditions, True)

    while True:
        comm.Recv(task, source=0, tag=77)
        if task[0] == TERMINATE:
            break
        else:
            settings['EbN0'] = task[1]
            sendBuffer[0] = worker(task[0], DicMtx, settings)
            comm.Isend(sendBuffer, dest=0, tag=77)

# WORKER ############################################################# WORKER #
def worker(task, DicMtx, settings):
    global activeRecvs
    recvId = task
    cmdString = 'errors = ef.'+activeRecvs[recvId]+'_ber_test(settings, DicMtx)'
    exec(cmdString)
    return errors

# GETPOLYS ######################################################### GETPOLYS #
def getPolys(m):
    polys = {
        3: list([[3, 1, 0], [3, 2, 0]]),
        5: list([[5, 2, 0], [5, 4, 3, 2, 0]]),
        7: list([[7, 6, 0], [7, 4, 0]]),
        10: list([[10, 3, 0], [10, 9, 8, 6, 3, 2, 0]])
        }
    
    initialConditions = sp.zeros((2,m))
    initialConditions[0,0]=1
    initialConditions[1,0]=1
    return polys[m], initialConditions

# MAIN ################################################################# MAIN #
def main(m, S, quanBits, kappa, noErrors, EbN0_Start, EbN0_End, expMask, noSymbolsPerRun):
    if rank == 0:
        # If we are on node 0 we will act as server
        server(m, S, quanBits, kappa, noErrors, EbN0_Start, EbN0_End, expMask, noSymbolsPerRun)
    else: 
        # "Workers" all nodes except the node with rank 0
        client()
